﻿/*
 * Author: G. Monz (DK7IO), 2013-11-22
 * This file is distributed without any warranty.
 * */

using System;
using System.Text;
using System.Xml;
using System.Collections.Generic;
using System.IO;
using Ionic.Zip;

namespace GM.Geodesy
{
	/// <summary>
	/// Creates KML files (Overlays for Google Earth).
	/// </summary>
	public static class KML
	{
		/// <summary>
		/// The output file format.
		/// </summary>
		public enum FileFormat
		{
			/// <summary>
			/// The KML file only.
			/// </summary>
			KmlOnly,

			/// <summary>
			/// The KMZ file only.
			/// </summary>
			KmzOnly,

			/// <summary>
			/// Both, the KML and the KMZ file.
			/// </summary>
			KmlAndKmz
		}

		/// <summary>
		/// Saves a KML file for a given <see cref="GeographicalPoint" /> with its waypoint projections.
		/// </summary>
		/// <param name="location">The <see cref="GeographicalPoint" /> instance.</param>
		/// <param name="waypoints">A list of <see cref="GeographicalPoint" /> instances (the waypoint projections).</param>
		/// <param name="fileNamePrefix">
		/// The full name of the KML file without extension ('.kml').
		/// The directory must exist with write permissions.
		/// </param>
		/// <param name="fileFormat">A <see cref="FileFormat" /> member.</param>
		/// <returns>The file name with extension.</returns>
		/// <exception cref="ArgumentException">If 'location' or the 'waypoints' instances are null.</exception>
		public static string SaveWaypointProjectionsKML(
			GeographicalPoint location, List<GeographicalPoint> waypoints,
			string fileNamePrefix,
			FileFormat fileFormat
		)
		{
			string identifier = "_" + Guid.NewGuid();
			fileNamePrefix += identifier;
			string effectiveFileName = fileNamePrefix + ".kml";

			XmlTextWriter xmlWriter = new XmlTextWriter(effectiveFileName, Encoding.UTF8);
			{
				xmlWriter.Formatting = Formatting.Indented;
				xmlWriter.IndentChar = '\t';
				xmlWriter.Indentation = 1;

				xmlWriter.WriteStartDocument();
				{
					xmlWriter.WriteStartElement("kml");
					{
						WriteKmlNamespaces(xmlWriter);

						xmlWriter.WriteStartElement("Document");
						{
							xmlWriter.WriteElementString("name", "Waypoint projections (by DK7IO)");

							WriteAboutSnippet(xmlWriter);

							WriteWaypointsStyle(xmlWriter);

							WriteFolder(xmlWriter, location, waypoints);
						}
						xmlWriter.WriteEndElement();
					}
					xmlWriter.WriteEndElement();
				}
				xmlWriter.WriteEndDocument();
			}
			xmlWriter.Close();

			string result = SaveKML(fileFormat, fileNamePrefix, effectiveFileName);

			return result;
		}

		static void WriteFolder(XmlTextWriter xmlWriter, GeographicalPoint location, List<GeographicalPoint> waypoints)
		{
			xmlWriter.WriteStartElement("Folder");
			{
				string name = location.Name;
				xmlWriter.WriteElementString("name", name);

				WriteWaypointsPlacemark(xmlWriter, location, name, "#Location");

				for (int i = 1; i < waypoints.Count; i++)
				{
					GeographicalPoint waypoint = waypoints[i];
					WriteWaypointsPlacemark(xmlWriter, waypoint, waypoint.Name, "#Waypoint");
				}

				xmlWriter.WriteStartElement("Placemark");
				{
					xmlWriter.WriteElementString("name", "Line of sight");

					xmlWriter.WriteStartElement("LineString");
					{
						xmlWriter.WriteElementString("tessellate", "1");

						xmlWriter.WriteElementString("coordinates", GetCoordinatesValue(waypoints));
					}
					xmlWriter.WriteEndElement();
				}
			}
			xmlWriter.WriteEndElement();
		}

		static void WriteKmlNamespaces(XmlTextWriter xmlWriter)
		{
			xmlWriter.WriteAttributeString("xmlns", "http://www.opengis.net/kml/2.2");
			xmlWriter.WriteAttributeString("xmlns:gx", "http://www.google.com/kml/ext/2.2");
			xmlWriter.WriteAttributeString("xmlns:kml", "http://www.opengis.net/kml/2.2");
			xmlWriter.WriteAttributeString("xmlns:atom", "http://www.w3.org/2005/Atom");
		}

		/// <summary>
		/// Saves a KML file for several 'Maidenhead Locators'.
		/// </summary>
		/// <param name="maidenheadLocators">A collection of 'Maidenhead Locators'.</param>
		/// <param name="superGrid">
		/// If True: Show next supergrid. If False: Don't show it.
		/// </param>
		/// <param name="allSuperGrids">
		/// If True: Show all supergrids. If False: Don't show them.
		/// </param>
		/// <param name="subGrids">
		/// If True: Show subgrids. If False: Don't show them.
		/// </param>
		/// <param name="fileNamePrefix">
		/// The full name of the KML file without extension ('.kml').
		/// The directory must exist with write permissions.
		/// </param>
		/// <param name="fileFormat">A <see cref="FileFormat" /> member.</param>
		/// <param name="KMLminLodPixels">
		/// The KML 'minLodPixels' value for the top region. Can be negative.
		/// A good value is 500.
		/// </param>
		/// <param name="KMLcolor_EvenPrecisions">
		/// The KML 'color' value for even locator precisions. Can be null. Example: ffff2222
		/// </param>
		/// <param name="KMLcolor_OddPrecisions">
		/// The KML 'color' value for odd locator precisions. Can be null. Example: ff0000ff
		/// </param>
		/// <returns>The file name with extension.</returns>
		/// <exception cref="ArgumentException">
		/// If the length of the locator text is null or not an even number.
		/// If the locator text contains invalid characters.
		/// </exception>
		public static string SaveMaidenheadLocatorsKML(
			ICollection<string> maidenheadLocators,
			bool superGrid, bool allSuperGrids,
			bool subGrids,
			string fileNamePrefix,
			FileFormat fileFormat,
			int KMLminLodPixels,
			string KMLcolor_EvenPrecisions, string KMLcolor_OddPrecisions
		)
		{
			//Corrections
			{
				KMLminLodPixels = Math.Max(KMLminLodPixels, 0);
			}

			string effectiveFileName = fileNamePrefix + ".kml";

			XmlTextWriter xmlWriter = new XmlTextWriter(effectiveFileName, Encoding.UTF8);
			{
				xmlWriter.Formatting = Formatting.Indented;
				xmlWriter.IndentChar = '\t';
				xmlWriter.Indentation = 1;

				xmlWriter.WriteStartDocument();
				{
					xmlWriter.WriteStartElement("kml");
					{
						WriteKmlNamespaces(xmlWriter);

						xmlWriter.WriteStartElement("Document");
						{
							xmlWriter.WriteElementString("name", "Maidenhead Locator (by DK7IO)");
							WriteAboutSnippet(xmlWriter);

							ICollection<int> precisions = new List<int>();

							foreach (string maidenheadLocatorsItem in maidenheadLocators)
							{
								string maidenheadLocator = (maidenheadLocatorsItem == null ? string.Empty : maidenheadLocator = maidenheadLocatorsItem.Trim());
								bool maidenheadLocatorIsEmpty = string.IsNullOrEmpty(maidenheadLocator);

								//Check arguments
								{
									if (!maidenheadLocatorIsEmpty)
									{
										//Dummy instance
										GeographicalPoint geographicalPoint = new GeographicalPoint(maidenheadLocator, PositionInRectangle.BottomLeft);
									}
								}

								int basePrecision = maidenheadLocator.Length / 2;
								WriteLocatorStyle(
									xmlWriter,
									KMLcolor_EvenPrecisions, KMLcolor_OddPrecisions,
									precisions, basePrecision
								);

								int superPrecision = basePrecision - 1;

								if (superPrecision < MaidenheadLocator.MinPrecision)
								{
									superGrid = allSuperGrids = false;
								}

								if (superGrid)
								{
									WriteLocatorStyle(
										xmlWriter,
										KMLcolor_EvenPrecisions, KMLcolor_OddPrecisions,
										precisions, superPrecision
									);

									if (allSuperGrids)
									{
										for (int precision = MaidenheadLocator.MinPrecision - 1; precision < superPrecision; precision++)
										{
											WriteLocatorStyle(
												xmlWriter,
												KMLcolor_EvenPrecisions, KMLcolor_OddPrecisions,
												precisions, precision
											);
										}
									}
								}

								int subPrecision = basePrecision + 1;

								if (basePrecision >= MaidenheadLocator.MaxPrecision)
								{
									subGrids = false;
								}
								else if (basePrecision < MaidenheadLocator.MinPrecision)
								{
									subGrids = true;
								}

								if (subGrids)
								{
									WriteLocatorStyle(
										xmlWriter,
										KMLcolor_EvenPrecisions, KMLcolor_OddPrecisions,
										precisions, subPrecision
									);
								}

								xmlWriter.WriteStartElement("Folder");
								{
									if (!maidenheadLocatorIsEmpty)
									{
										xmlWriter.WriteElementString("name", maidenheadLocator);

										WriteLocatorRegion(xmlWriter, maidenheadLocator, KMLminLodPixels);

										WriteLocatorPlacemark(xmlWriter, maidenheadLocator, basePrecision, true, (int)Math.Sqrt(KMLminLodPixels));
									}

									//Supergrids
									if (superGrid)
									{
										int minPrecision = (
											allSuperGrids ?
											MaidenheadLocator.MinPrecision :
											superPrecision
										);
										GeographicalPoint geographicalPoint = new GeographicalPoint(maidenheadLocator, PositionInRectangle.MiddleMiddle);

										for (int precision = minPrecision; precision < basePrecision; precision++)
										{
											string superLocator = geographicalPoint.GetMaidenheadLocator(true, precision);
											WriteLocatorPlacemark(xmlWriter, superLocator, precision, true, 0);
										}
									}

									//Subgrids
									if (subGrids)
									{
										xmlWriter.WriteStartElement("Folder");
										{
											xmlWriter.WriteElementString("name", maidenheadLocator + "..");

											List<string> subLocators = MaidenheadLocator.GetSubGrids(maidenheadLocator, MaidenheadLocator.SubGridsFilter.All, true, false);

											foreach (string subLocator in subLocators)
											{
												WriteLocatorPlacemark(xmlWriter, subLocator, subPrecision, false, 0);
											}
										}
										xmlWriter.WriteEndElement();
									}
								}
								xmlWriter.WriteEndElement();
							}
						}
						xmlWriter.WriteEndElement();
					}
					xmlWriter.WriteEndElement();
				}
				xmlWriter.WriteEndDocument();
			}
			xmlWriter.Close();

			string result = SaveKML(fileFormat, fileNamePrefix, effectiveFileName);

			return result;
		}

		static void WriteAboutSnippet(XmlTextWriter xmlWriter)
		{
			xmlWriter.WriteStartElement("Snippet");
			{
				string url = "http://dk7io.darc.de/programmierung/GM.Geodesy/";
				string info =
					"Created with " +
					"<A href=" +
						"\"" + url + "\"" + ">" +
						typeof(KML).Namespace +
					"</A>"
				;
				xmlWriter.WriteCData(info);
			}
			xmlWriter.WriteEndElement();
		}

		static string SaveKML(FileFormat fileFormat, string fileNamePrefix, string effectiveFileName)
		{
			string result;

			if (
				fileFormat == FileFormat.KmzOnly ||
				fileFormat == FileFormat.KmlAndKmz
			)
			{
				result = SaveKMZ(fileNamePrefix, effectiveFileName, fileFormat);
			}
			else
			{
				result = effectiveFileName;
			}

			return result;
		}

		static string SaveKMZ(string fileName, string kmlFile, FileFormat fileFormat)
		{
			string result = fileName + ".kmz";

			using (ZipFile zip = new ZipFile())
			{
				zip.AddFile(kmlFile);
				zip.Save(result);
			}

			if (fileFormat == FileFormat.KmzOnly)
			{
				new FileInfo(kmlFile).Delete();
			}

			return result;
		}

		static void WriteLocatorStyle(
			XmlTextWriter xmlWriter,
			string KMLcolor_EvenPrecisions, string KMLcolor_OddPrecisions,
			ICollection<int> precisions, int precision
		)
		{
			if (precision >= 0)
			{
				if (!precisions.Contains(precision))
				{
					string color = (
						(precision % 2 == 0) ?
						KMLcolor_EvenPrecisions :
						KMLcolor_OddPrecisions
					);

					xmlWriter.WriteStartElement("Style");
					{
						xmlWriter.WriteAttributeString("id", "Precision" + precision.ToString());

						xmlWriter.WriteStartElement("IconStyle");
						{
							xmlWriter.WriteStartElement("Icon");
							{
							}
							xmlWriter.WriteEndElement();
						}
						xmlWriter.WriteEndElement();

						xmlWriter.WriteStartElement("LabelStyle");
						{
							xmlWriter.WriteElementString("color", color);
						}
						xmlWriter.WriteEndElement();

						xmlWriter.WriteStartElement("LineStyle");
						{
							xmlWriter.WriteElementString("color", color);
						}
						xmlWriter.WriteEndElement();
					}
					xmlWriter.WriteEndElement();

					precisions.Add(precision);
				}
			}
		}

		static void WriteWaypointsStyle(
			XmlTextWriter xmlWriter
		)
		{
			WriteWaypointsStyle(xmlWriter, "Location", "http://maps.google.com/mapfiles/kml/shapes/square.png");
			WriteWaypointsStyle(xmlWriter, "Waypoint", "http://maps.google.com/mapfiles/kml/shapes/open-diamond.png");
		}

		static void WriteWaypointsStyle(
			XmlTextWriter xmlWriter,
			string id, string iconURL
		)
		{
			xmlWriter.WriteStartElement("Style");
			{
				xmlWriter.WriteAttributeString("id", id + "Normal");

				xmlWriter.WriteStartElement("IconStyle");
				{
					xmlWriter.WriteElementString("scale", "1.2");

					xmlWriter.WriteStartElement("Icon");
					{
						xmlWriter.WriteElementString("href", iconURL);
					}
					xmlWriter.WriteEndElement();
				}
				xmlWriter.WriteEndElement();

				xmlWriter.WriteStartElement("ListStyle");
				{
				}
				xmlWriter.WriteEndElement();
			}
			xmlWriter.WriteEndElement();


			xmlWriter.WriteStartElement("Style");
			{
				xmlWriter.WriteAttributeString("id", id + "Highlight");

				xmlWriter.WriteStartElement("IconStyle");
				{
					xmlWriter.WriteElementString("scale", "1.4");

					xmlWriter.WriteStartElement("Icon");
					{
						xmlWriter.WriteElementString("href", iconURL);
					}
					xmlWriter.WriteEndElement();
				}
				xmlWriter.WriteEndElement();

				xmlWriter.WriteStartElement("ListStyle");
				{
				}
				xmlWriter.WriteEndElement();
			}
			xmlWriter.WriteEndElement();

			xmlWriter.WriteStartElement("StyleMap");
			{
				xmlWriter.WriteAttributeString("id", id);

				xmlWriter.WriteStartElement("Pair");
				{
					xmlWriter.WriteElementString("key", "normal");
					xmlWriter.WriteElementString("styleUrl", "#" + id + "Normal");
				}
				xmlWriter.WriteEndElement();

				xmlWriter.WriteStartElement("Pair");
				{
					xmlWriter.WriteElementString("key", "highlight");
					xmlWriter.WriteElementString("styleUrl", "#" + id + "Highlight");
				}
				xmlWriter.WriteEndElement();
			}
			xmlWriter.WriteEndElement();
		}

		static void WriteLocatorRegion(XmlTextWriter xmlWriter, string locator, int KMLminLodPixels)
		{
			xmlWriter.WriteStartElement("Region");
			{
				xmlWriter.WriteStartElement("LatLonAltBox");
				{
					xmlWriter.WriteElementString("north", Format(new GeographicalPoint(locator, PositionInRectangle.TopMiddle).Latitude));
					xmlWriter.WriteElementString("south", Format(new GeographicalPoint(locator, PositionInRectangle.BottomMiddle).Latitude));
					xmlWriter.WriteElementString("east", Format(new GeographicalPoint(locator, PositionInRectangle.MiddleRight).Longitude));
					xmlWriter.WriteElementString("west", Format(new GeographicalPoint(locator, PositionInRectangle.MiddleLeft).Longitude));
				}
				xmlWriter.WriteEndElement();

				xmlWriter.WriteStartElement("Lod");
				{
					xmlWriter.WriteElementString("minLodPixels", KMLminLodPixels.ToString());
				}
				xmlWriter.WriteEndElement();
			}
			xmlWriter.WriteEndElement();
		}

		static void WriteLocatorPlacemark(XmlTextWriter xmlWriter, string locator, int precision, bool writeRegion, int regionMinLodPixels)
		{
			xmlWriter.WriteStartElement("Placemark");
			{
				xmlWriter.WriteElementString("name", locator);
				xmlWriter.WriteElementString("styleUrl", "#" + "Precision" + precision.ToString());

				if (writeRegion)
				{
					WriteLocatorRegion(xmlWriter, locator, regionMinLodPixels);
				}

				xmlWriter.WriteStartElement("MultiGeometry");
				{
					xmlWriter.WriteStartElement("Point");
					{
						GeographicalPoint center = new GeographicalPoint(locator, PositionInRectangle.MiddleMiddle);
						xmlWriter.WriteElementString("coordinates", GetCoordinatesValue(GeographicalPoint.GetNewList(center)));
					}
					xmlWriter.WriteEndElement();

					xmlWriter.WriteStartElement("LineString");
					{
						xmlWriter.WriteElementString("tessellate", "1");

						List<GeographicalPoint> points = GetOutlinePoints(locator, true);
						xmlWriter.WriteElementString("coordinates", GetCoordinatesValue(points));
					}
					xmlWriter.WriteEndElement();
				}
				xmlWriter.WriteEndElement();
			}
			xmlWriter.WriteEndElement();
		}

		static void WriteWaypointsPlacemark(XmlTextWriter xmlWriter, GeographicalPoint geographicalPoint, string name, string styleUrl)
		{
			xmlWriter.WriteStartElement("Placemark");
			{
				xmlWriter.WriteElementString("name", name);
				xmlWriter.WriteElementString("styleUrl", styleUrl);

				xmlWriter.WriteStartElement("Point");
				{
					xmlWriter.WriteElementString("coordinates", GetCoordinatesValue(GeographicalPoint.GetNewList(geographicalPoint)));
				}
				xmlWriter.WriteEndElement();
			}
			xmlWriter.WriteEndElement();
		}

		static List<GeographicalPoint> GetOutlinePoints(string baseLocator, bool fromSubLocators)
		{
			List<GeographicalPoint> points = new List<GeographicalPoint>();
			{
				if (fromSubLocators)
				{
					List<string> bottomSubLocators = MaidenheadLocator.GetSubGrids(baseLocator, MaidenheadLocator.SubGridsFilter.Bottom, true, false);
					List<string> rightSubLocators = MaidenheadLocator.GetSubGrids(baseLocator, MaidenheadLocator.SubGridsFilter.Right, true, false);
					List<string> topSubLocators = MaidenheadLocator.GetSubGrids(baseLocator, MaidenheadLocator.SubGridsFilter.Top, true, true);
					List<string> leftSubLocators = MaidenheadLocator.GetSubGrids(baseLocator, MaidenheadLocator.SubGridsFilter.Left, true, true);

					//Bottom
					{
						foreach (string subLocator in bottomSubLocators)
						{
							points.Add(new GeographicalPoint(subLocator, PositionInRectangle.BottomLeft));
						}
					}

					//Right
					{
						foreach (string subLocator in rightSubLocators)
						{
							points.Add(new GeographicalPoint(subLocator, PositionInRectangle.BottomRight));
						}
					}

					//Top
					{
						foreach (string subLocator in topSubLocators)
						{
							points.Add(new GeographicalPoint(subLocator, PositionInRectangle.TopRight));
						}
					}

					//Left
					{
						foreach (string subLocator in leftSubLocators)
						{
							points.Add(new GeographicalPoint(subLocator, PositionInRectangle.TopLeft));
						}
					}

					//Last point towards start point
					{
						if (bottomSubLocators.Count > 0)
						{
							points.Add(new GeographicalPoint(bottomSubLocators[0], PositionInRectangle.BottomLeft));
						}
					}
				}
				else
				{
					points.Add(new GeographicalPoint(baseLocator, PositionInRectangle.BottomLeft));
					points.Add(new GeographicalPoint(baseLocator, PositionInRectangle.BottomRight));
					points.Add(new GeographicalPoint(baseLocator, PositionInRectangle.TopRight));
					points.Add(new GeographicalPoint(baseLocator, PositionInRectangle.TopLeft));
					points.Add(new GeographicalPoint(baseLocator, PositionInRectangle.BottomLeft));
				}
			}

			return points;
		}

		static string Format(double value)
		{
			string result = value.ToString();
			result = result.Replace(',', '.');

			return result;
		}

		static string Format(GeographicalPoint geographicalPoint)
		{
			string result =
				Format(geographicalPoint.Longitude) +
				"," +
				Format(geographicalPoint.Latitude)
			;

			return result;
		}

		static string GetCoordinatesValue(List<GeographicalPoint> geographicalPoints)
		{
			StringBuilder text = new StringBuilder();
			{
				int count = geographicalPoints.Count;
				int i = 0;

				foreach (GeographicalPoint geographicalPoint in geographicalPoints)
				{
					text.Append(Format(geographicalPoint))
						.Append(",0")
					;

					if (i < count - 1)
					{
						text.Append(" ");
					}

					i++;
				}
			}
			string result = text.ToString();

			return result;
		}
	}
}