/*
 * Author: G. Monz (DK7IO), 2013-11-22
 * This file is distributed without any warranty.
 * */

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Drawing;
using System.Windows.Forms;
using GM.Geodesy;

namespace QthManager
{
	public partial class MainForm : Form
	{
		class ListControlItem
		{
			object _data;
			public object Data
			{
				get
				{
					return _data;
				}
				private set
				{
					_data = value;
				}
			}

			string _displayText;
			public string DisplayText
			{
				get
				{
					return _displayText;
				}
				private set
				{
					_displayText = value;
				}
			}

			internal ListControlItem(object data, string displayText)
			{
				Data = data;
				DisplayText = displayText;
			}
		}

		const int _maxCsvDecimalPlaces = 7;

		KML.FileFormat KMLFileFormat
		{
			get
			{
				KML.FileFormat result = KML.FileFormat.KmzOnly;

				bool found = false;
				{
					Control.ControlCollection controls = groupBoxKmlFile.Controls;

					foreach (Control control in controls)
					{
						RadioButton radioButton = control as RadioButton;

						if (radioButton != null)
						{
							if (radioButton.Checked)
							{
								string tag = control.Tag as string;

								if (tag != null)
								{
									TypeConverter converter = TypeDescriptor.GetConverter(typeof(KML.FileFormat));
									result = (KML.FileFormat)converter.ConvertFromInvariantString(tag);
									found = true;
									break;
								}
							}
						}
					}
				}

				if (found)
				{
					return result;
				}
				else
				{
					throw new InvalidOperationException("No valid RadioButton checked.");
				}
			}
		}

		AngleFormat AngleFormat
		{
			get
			{
				ListControlItem listControlItem = comboBoxCsvAngleFormat.SelectedItem as ListControlItem;
				AngleFormat result = (AngleFormat)listControlItem.Data;

				return result;
			}
		}

		public MainForm()
		{
			InitializeComponent();

			{
				Text = "QTH Manager (by DK7IO, 2013-11-22)";

				radioButtonLocatorPositionInGridMiddleMiddle1.Checked = radioButtonLocatorPositionInGridMiddleMiddle2.Checked = true;

				radioButtonOutputKMZ.Checked = true;

				textBoxminLodPixels.Text = (500).ToString();

				textBoxcolor_EvenPrecisions.Text = "ffff2222";
				textBoxcolor_OddPrecisions.Text = "ff0000ff";

				ResetState1();
				ResetState2();

				textBoxOutputPath.Text = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments);

				_saveCsvContextMenu1.ItemClicked += new ToolStripItemClickedEventHandler(SaveCsvContextMenu1_ItemClicked);
				_saveCsvContextMenu2.ItemClicked += new ToolStripItemClickedEventHandler(SaveCsvContextMenu2_ItemClicked);

				_saveKmlContextMenu1.ItemClicked += new ToolStripItemClickedEventHandler(SaveKmlContextMenu1_ItemClicked);
				_saveKmlContextMenu2.ItemClicked += new ToolStripItemClickedEventHandler(SaveKmlContextMenu2_ItemClicked);

				textBoxWaypointsAzimuth1.Text = textBoxWaypointsAzimuth2.Text = (270).ToString();
				textBoxWaypointsCount1.Text = textBoxWaypointsCount2.Text = (20).ToString();
				textBoxWaypointsKm1.Text = textBoxWaypointsKm2.Text = (0.5).ToString();

				textBoxLocator1.Text = textBoxLocator2.Text = null;
				LocatorTextBoxChanged(
					textBoxLocator1.Text, buttonLocatorEnter1,
					buttonSaveKml1, buttonSaveCsv1, buttonWaypoints1
				);
				LocatorTextBoxChanged(
					textBoxLocator2.Text, buttonLocatorEnter2,
					buttonSaveKml2, buttonSaveCsv2, buttonWaypoints2
				);

				textBoxCsvDecimalPlaces.Text = (1).ToString();
				labelCsvDecimalPlaces.Text = "Decimal places (0..." + _maxCsvDecimalPlaces.ToString() + ")";
				RefreshComboBoxCsvAngleFormat(false);

				checkBoxSuperGrid.Checked = false;
				checkBoxAllSuperGrids.Checked = true;
				SupergridCheckBoxChanged();

				checkBoxSubGrids.Checked = false;
			}
		}

		void RefreshComboBoxCsvAngleFormat(bool keepSelectedIndex)
		{
			List<ListControlItem> angleFormatItems = new List<ListControlItem>();
			{
				GeographicalPoint geographicalPoint = new GeographicalPoint(
					-4 + Math.Sqrt(2),
					+7 + Math.Sqrt(3)
				);

				int decimalPlaces;
				bool decimalPlacesOK = int.TryParse(textBoxCsvDecimalPlaces.Text, out decimalPlaces);

				if (decimalPlacesOK)
				{
					foreach (AngleFormat angleFormat in Enum.GetValues(typeof(AngleFormat)))
					{
						string displayText = geographicalPoint.ToString(angleFormat, decimalPlaces, " / ", false);
						ListControlItem listControlItem = new ListControlItem(angleFormat, displayText);
						angleFormatItems.Add(listControlItem);
					}
				}
			}
			int selectedIndex = (keepSelectedIndex ? comboBoxCsvAngleFormat.SelectedIndex : angleFormatItems.Count - 1);
			comboBoxCsvAngleFormat.DataSource = angleFormatItems;
			comboBoxCsvAngleFormat.SelectedIndex = selectedIndex;
		}

		void ShowSaveCsvContextMenu(MouseEventArgs mouseEventArgs, Control control, ContextMenuStrip contextMenu, string locator)
		{
			locator = locator.Trim();

			ToolStripItemCollection items = contextMenu.Items;
			{
				items.Clear();

				{
					ToolStripMenuItem item = new ToolStripMenuItem("Save CSV for '" + locator + "'");
					{
						item.Tag = new List<string>(new string[] { locator });
						item.Image = Properties.Resources.go;
					}
					items.Add(item);
				}

				if (locator.Length / 2 < MaidenheadLocator.MaxPrecision)
				{
					List<string> subLocators = MaidenheadLocator.GetSubGrids(locator, MaidenheadLocator.SubGridsFilter.All, checkBoxSmallLettersForSubsquares.Checked, false);
					string firstSubLocator = subLocators[0];
					string lastSubLocator = subLocators[subLocators.Count - 1];

					ToolStripMenuItem item = new ToolStripMenuItem("Save CSV for '" + firstSubLocator + "' ... '" + lastSubLocator + "' (and do not open files)");
					{
						item.Tag = subLocators;
						item.Image = Properties.Resources.fast_forward;
					}
					items.Add(item);
				}
			}

			contextMenu.Show(control, mouseEventArgs.X, mouseEventArgs.Y);
		}

		void ShowSaveKmlContextMenu(MouseEventArgs mouseEventArgs, Control control, ContextMenuStrip contextMenu, string locator)
		{
			locator = locator.Trim();

			ToolStripItemCollection items = contextMenu.Items;
			{
				items.Clear();

				{
					ToolStripMenuItem item = new ToolStripMenuItem("Save KML for '" + locator + "'");
					{
						item.Tag = new List<string>(new string[] { locator });
						item.Image = Properties.Resources.go;
					}
					items.Add(item);
				}

				if (locator.Length / 2 < MaidenheadLocator.MaxPrecision)
				{
					List<string> subLocators = MaidenheadLocator.GetSubGrids(locator, MaidenheadLocator.SubGridsFilter.All, checkBoxSmallLettersForSubsquares.Checked, false);
					string firstSubLocator = subLocators[0];
					string lastSubLocator = subLocators[subLocators.Count - 1];

					ToolStripMenuItem item = new ToolStripMenuItem("Save KML for '" + firstSubLocator + "' ... '" + lastSubLocator + "' (and do not open files)");
					{
						item.Tag = subLocators;
						item.Image = Properties.Resources.fast_forward;
					}
					items.Add(item);
				}
			}

			contextMenu.Show(control, mouseEventArgs.X, mouseEventArgs.Y);
		}

		void SaveCsvContextMenu_ItemClicked(ToolStripItemClickedEventArgs toolStripItemClickedEventArgs, TextBox textBoxCurrentLocator)
		{
			ToolStripItem item = toolStripItemClickedEventArgs.ClickedItem;

			if (item != null)
			{
				List<string> locators = item.Tag as List<string>;

				if (locators != null)
				{
					Application.DoEvents();
					SetBusyCursor();
					bool checkOpenFile = (locators.Count == 1);

					foreach (string locator in locators)
					{
						SaveCsv(locator, textBoxCurrentLocator, checkOpenFile);
					}
				}
			}
		}

		void SaveKmlContextMenu_ItemClicked(ToolStripItemClickedEventArgs toolStripItemClickedEventArgs, TextBox textBoxCurrentLocator)
		{
			ToolStripItem item = toolStripItemClickedEventArgs.ClickedItem;

			if (item != null)
			{
				List<string> locators = item.Tag as List<string>;

				if (locators != null)
				{
					Application.DoEvents();
					SetBusyCursor();
					bool checkOpenFile = (locators.Count == 1);

					foreach (string locator in locators)
					{
						SaveKml(locator, textBoxCurrentLocator, checkOpenFile);
					}
				}
			}
		}

		void textBoxCsvDecimalPlaces_Validating(object sender, CancelEventArgs e)
		{
			int decimalPlaces;
			bool success = int.TryParse(textBoxCsvDecimalPlaces.Text, out decimalPlaces);

			e.Cancel =
				!success ||
				!(decimalPlaces >= 0 && decimalPlaces <= _maxCsvDecimalPlaces)
			;
		}

		void textBoxCsvDecimalPlaces_Validated(object sender, EventArgs e)
		{
			RefreshComboBoxCsvAngleFormat(true);
		}

		void buttonOutputPathSelect_Click(object sender, EventArgs e)
		{
			DialogResult result = folderBrowserDialogOutputPath.ShowDialog();

			if (result == DialogResult.OK)
			{
				textBoxOutputPath.Text = folderBrowserDialogOutputPath.SelectedPath;
			}
		}

		void checkBoxSupergrid_CheckedChanged(object sender, EventArgs e)
		{
			SupergridCheckBoxChanged();
		}

		void SupergridCheckBoxChanged()
		{
			checkBoxAllSuperGrids.Enabled = checkBoxSuperGrid.Checked;
		}

		void LocatorTextBoxChanged(
			string locator, Button buttonLocatorEnter,
			Button buttonSaveKml, Button buttonSaveCsv, Button buttonWaypoints
		)
		{
			bool enableButtons = false;

			try
			{
				locator = (
					string.IsNullOrEmpty(locator) ?
					string.Empty :
					locator.Trim()
				);

				toolTip.SetToolTip(
					buttonLocatorEnter,
					"Apply locator" + Environment.NewLine + locator
				);

				{
					string postfix =
						" file" + Environment.NewLine +
						"for locator" + Environment.NewLine +
						locator + Environment.NewLine + Environment.NewLine +
						"or: right click for more options"
					;

					toolTip.SetToolTip(buttonSaveCsv, "Save locator CSV" + postfix);
					toolTip.SetToolTip(buttonSaveKml, "Save locator KML" + postfix);
				}

				{
					string postfix =
						" file" + Environment.NewLine +
						"for locator" + Environment.NewLine +
						locator
					;

					toolTip.SetToolTip(buttonWaypoints, "Save waypoints KML" + postfix);
				}

				if (string.IsNullOrEmpty(locator))
				{
					enableButtons = true;
				}
				else
				{
					GeographicalPoint dummy = new GeographicalPoint(locator, PositionInRectangle.BottomLeft);
					enableButtons = true;
				}
			}
			catch
			{
			}

			buttonSaveKml.Enabled = buttonSaveCsv.Enabled = enableButtons;

			if (string.IsNullOrEmpty(locator))
			{
				buttonWaypoints.Enabled = false;
			}
			else
			{
				buttonWaypoints.Enabled = enableButtons;
			}
		}

		void ResetLocator(TextBox textBoxLocator, ListBox listBoxLocatorPrecisions)
		{
			textBoxLocator.Text = null;
			listBoxLocatorPrecisions.Items.Clear();
		}

		PositionInRectangle GetLocatorPositionInGrid(GroupBox groupBox)
		{
			PositionInRectangle result = PositionInRectangle.MiddleMiddle;

			bool found = false;
			{
				Control.ControlCollection controls = groupBox.Controls;

				foreach (Control control in controls)
				{
					RadioButton radioButton = control as RadioButton;

					if (radioButton != null)
					{
						if (radioButton.Checked)
						{
							string tag = control.Tag as string;

							if (tag != null)
							{
								TypeConverter converter = TypeDescriptor.GetConverter(typeof(PositionInRectangle));
								result = (PositionInRectangle)converter.ConvertFromInvariantString(tag);
								found = true;
								break;
							}
						}
					}
				}
			}

			if (found)
			{
				return result;
			}
			else
			{
				throw new InvalidOperationException("No valid RadioButton checked.");
			}
		}

		static int GetListBoxLocatorPrecisionsIndex(string locator)
		{
			int result = Math.Min(
				MaidenheadLocator.MaxPrecision - 1,
				locator.Length / 2 - 1
			);

			return result;
		}

		static string Get0Value(TextBox textBox)
		{
			string result = "0";
			{
				string text = textBox.Text;

				if (!string.IsNullOrEmpty(text))
				{
					result = text;
				}
			}

			return result;
		}

		void SaveCsv(string locator, TextBox textBoxCurrentLocator, bool checkOpenFile)
		{
			string fileName = "Locator_" + locator;
			string fullFileName = GetNormalizedPathName(textBoxOutputPath.Text, true) + fileName + ".csv";

			textBoxCurrentLocator.Text = fileName;
			Refresh();
			{
				int decimalPlaces;
				bool decimalPlacesOK = int.TryParse(textBoxCsvDecimalPlaces.Text, out decimalPlaces);

				if (decimalPlacesOK)
				{
					AngleFormat angleFormat = AngleFormat;

					CSV.SaveCSV(
						locator,
						fullFileName,
						checkBoxCsvUtf8.Checked,
						angleFormat,
						decimalPlaces
					);

					if (
						checkOpenFile &&
						checkBoxOpenKml.Checked
					)
					{
						StartProcess(fullFileName);
					}
				}
				else
				{
					throw new InvalidOperationException(labelCsvDecimalPlaces.Text + ": Invalid input");
				}
			}
			textBoxCurrentLocator.Text = null;
			Refresh();
		}

		void SaveKml(string locator, TextBox textBoxCurrentLocator, bool checkOpenFile)
		{
			int minLodPixels;
			bool minLodPixelsOK = int.TryParse(textBoxminLodPixels.Text, out minLodPixels);

			if (minLodPixelsOK)
			{
				KML.FileFormat fileFormat = KMLFileFormat;

				string fileNamePrefix = "Locator_" + locator;
				string fullFileNamePrefix = GetNormalizedPathName(textBoxOutputPath.Text, true) + fileNamePrefix;

				textBoxCurrentLocator.Text = fileNamePrefix;
				Refresh();
				{
					ICollection<string> locators = new List<string>();
					locators.Add(locator);

					string effectiveFileName = KML.SaveMaidenheadLocatorsKML(
						locators,
						checkBoxSuperGrid.Checked, checkBoxAllSuperGrids.Checked,
						checkBoxSubGrids.Checked,
						fullFileNamePrefix,
						fileFormat,
						minLodPixels,
						textBoxcolor_EvenPrecisions.Text, textBoxcolor_OddPrecisions.Text
					);

					if (
						checkOpenFile &&
						checkBoxOpenKml.Checked
					)
					{
						StartProcess(effectiveFileName);
					}
				}
				textBoxCurrentLocator.Text = null;
				Refresh();
			}
			else
			{
				throw new InvalidOperationException(labelminLodPixels.Text + ": Invalid input");
			}
		}

		void SaveWaypoints(
			GeographicalPoint geographicalPoint, string locator,
			TextBox textBoxWaypointsAzimuth, TextBox textBoxWaypointsCount, TextBox textBoxWaypointsKm
		)
		{
			double waypointsAzimuth = double.Parse(textBoxWaypointsAzimuth.Text);
			int waypointsCount = int.Parse(textBoxWaypointsCount.Text);
			double waypointsKm = double.Parse(textBoxWaypointsKm.Text);

			AngleFormat angleFormat = AngleFormat;

			int decimalPlaces = int.Parse(textBoxCsvDecimalPlaces.Text);

			bool forceAsciiCharacters = !checkBoxCsvUtf8.Checked;

			string separator = " / ";

			geographicalPoint.Name =
				geographicalPoint.ToString(angleFormat, decimalPlaces, separator, forceAsciiCharacters)
			;

			List<GeographicalPoint> waypoints = new List<GeographicalPoint>();
			{
				for (int i = 0; i <= waypointsCount; i++)
				{
					double km = waypointsKm * i;

					GeographicalPoint waypoint = geographicalPoint.GetWaypointProjection(waypointsAzimuth, km);
					{
						waypoint.Name =
							km.ToString() + " km" +
							": " +
							waypoint.ToString(angleFormat, decimalPlaces, separator, forceAsciiCharacters)
						;
					}
					waypoints.Add(waypoint);
				}
			}

			KML.FileFormat fileFormat = KMLFileFormat;

			string fileNamePrefix = "Waypoints_" + locator;
			string fullFileNamePrefix = GetNormalizedPathName(textBoxOutputPath.Text, true) + fileNamePrefix;

			string effectiveFileName = KML.SaveWaypointProjectionsKML(
				geographicalPoint, waypoints,
				fullFileNamePrefix,
				fileFormat
			);

			if (checkBoxOpenKml.Checked)
			{
				StartProcess(effectiveFileName);
			}
		}

		void StartProcess(string fileName)
		{
			Process.Start(fileName);
		}

		void SetState(TextBox textBoxLocatorState, Exception exception)
		{
			textBoxLocatorState.Tag = exception;
			textBoxLocatorState.Text = exception.Message;

			textBoxLocatorState.BackColor = SystemColors.Info;
			textBoxLocatorState.ForeColor = SystemColors.InfoText;

			textBoxLocatorState.Show();

			toolTip.SetToolTip(textBoxLocatorState, "Double click to display full error message");

			SetDefaultCursor();
		}

		void ResetState(TextBox textBoxLocatorState, LinkLabel linkLabelGeoHack, Label labelGeoHack, TextBox textBoxCurrentLocator)
		{
			textBoxLocatorState.Text = null;
			textBoxLocatorState.Tag = null;

			textBoxLocatorState.BackColor = SystemColors.Control;
			textBoxLocatorState.ForeColor = SystemColors.WindowText;

			textBoxLocatorState.Hide();

			ResetMapProvider(linkLabelGeoHack, labelGeoHack);

			textBoxCurrentLocator.Text = null;
		}

		void ShowException(TextBox textBoxLocatorState)
		{
			Exception exception = textBoxLocatorState.Tag as Exception;

			if (exception != null)
			{
				textBoxLocatorState.Text = exception.ToString();
			}

			toolTip.SetToolTip(textBoxLocatorState, null);
		}

		void ButtonLocatorPrecisionEnter(Button buttonLocatorPrecisionEnter, TextBox textBoxLocator)
		{
			string text = null;
			{
				object tag = buttonLocatorPrecisionEnter.Tag;

				if (tag != null)
				{
					text = tag as string;
				}
			}
			textBoxLocator.Text = text;
		}

		void ListBoxLocatorPrecisionsSelectedIndexChanged(ListBox listBoxLocatorPrecisions, Button buttonLocatorPrecisionEnter)
		{
			string locator = listBoxLocatorPrecisions.SelectedItem as string;
			buttonLocatorPrecisionEnter.Tag = locator;
			toolTip.SetToolTip(buttonLocatorPrecisionEnter, "Pick locator" + Environment.NewLine + locator);
		}

		void SetMapProvider(LinkLabel linkLabel, Label label, GeographicalPoint geographicalPoint)
		{
			string url = GeographicalPoint.GetURL(geographicalPoint, MapProvider.GeoHack);
			SetText(url, linkLabel, label);
		}

		void ResetMapProvider(LinkLabel linkLabel, Label label)
		{
			SetText(null, linkLabel, label);
		}

		void SetText(string text, LinkLabel linkLabel, Label label)
		{
			linkLabel.Text = text;
			toolTip.SetToolTip(linkLabel, text);

			label.Visible = !string.IsNullOrEmpty(text);
		}

		static string GetFormatted(double value, int decimalPlaces)
		{
			string format = "{0:0." + new string('0', decimalPlaces) + "}";
			string result = string.Format(format, value);

			return result;
		}

		void GeographicalPointUpdated(
			GeographicalPoint geographicalPoint,
			TextBox textBoxDND, TextBox textBoxDED,
			TextBox textBoxDMND, TextBox textBoxDMNM, TextBox textBoxDMED, TextBox textBoxDMEM,
			TextBox textBoxDMSND, TextBox textBoxDMSNM, TextBox textBoxDMSNS, TextBox textBoxDMSED, TextBox textBoxDMSEM, TextBox textBoxDMSES
		)
		{
			//D
			{
				int decimalPlaces = _maxCsvDecimalPlaces + 2;

				{
					bool isNegative;
					double degrees;
					Utilities.DegreesToD(geographicalPoint.Latitude, out isNegative, out degrees);

					textBoxDND.Text =
						GeographicalPoint.GetSignSymbol(isNegative) +
						GetFormatted(Math.Abs(geographicalPoint.Latitude), decimalPlaces)
					;
				}
				{
					bool isNegative;
					double degrees;
					Utilities.DegreesToD(geographicalPoint.Longitude, out isNegative, out degrees);

					textBoxDED.Text =
						GeographicalPoint.GetSignSymbol(isNegative) +
						GetFormatted(Math.Abs(geographicalPoint.Longitude), decimalPlaces)
					;
				}
			}

			//DM
			{
				int decimalPlaces = _maxCsvDecimalPlaces + 2;

				{
					bool isNegative;
					int degrees;
					double minutes;
					Utilities.DegreesToDM(geographicalPoint.Latitude, out isNegative, out degrees, out minutes);

					Utilities.Round(decimalPlaces, ref degrees, ref minutes);

					textBoxDMND.Text =
						GeographicalPoint.GetSignSymbol(isNegative) +
						degrees.ToString()
					;
					textBoxDMNM.Text = GetFormatted(minutes, decimalPlaces);
				}
				{
					bool isNegative;
					int degrees;
					double minutes;
					Utilities.DegreesToDM(geographicalPoint.Longitude, out isNegative, out degrees, out minutes);

					Utilities.Round(decimalPlaces, ref degrees, ref minutes);

					textBoxDMED.Text =
						GeographicalPoint.GetSignSymbol(isNegative) +
						degrees.ToString()
					;
					textBoxDMEM.Text = GetFormatted(minutes, decimalPlaces);
				}
			}

			//DMS
			{
				int decimalPlaces = _maxCsvDecimalPlaces;

				{
					bool isNegative;
					int degrees;
					int minutes;
					double seconds;
					Utilities.DegreesToDMS(geographicalPoint.Latitude, out isNegative, out degrees, out minutes, out seconds);

					Utilities.Round(decimalPlaces, ref degrees, ref minutes, ref seconds);

					textBoxDMSND.Text =
						GeographicalPoint.GetSignSymbol(isNegative) +
						degrees.ToString()
					;
					textBoxDMSNM.Text = minutes.ToString();
					textBoxDMSNS.Text = GetFormatted(seconds, decimalPlaces);
				}
				{
					bool isNegative;
					int degrees;
					int minutes;
					double seconds;
					Utilities.DegreesToDMS(geographicalPoint.Longitude, out isNegative, out degrees, out minutes, out seconds);

					Utilities.Round(decimalPlaces, ref degrees, ref minutes, ref seconds);

					textBoxDMSED.Text =
						GeographicalPoint.GetSignSymbol(isNegative) +
						degrees.ToString()
					;
					textBoxDMSEM.Text = minutes.ToString();
					textBoxDMSES.Text = GetFormatted(seconds, decimalPlaces);
				}
			}

			GeographicalPointsUpdated();
		}

		void GeographicalPointsUpdated()
		{
			int decimalPlaces = 6;

			if (GeographicalPoint1 != null && GeographicalPoint2 != null)
			{
				{
					double distanceKM = GeographicalPoint1.GetDistance(GeographicalPoint2);
					double distanceStatuteMiles = distanceKM / GeographicalPoint.StatuteMilesToKilometres;
					double distanceNauticalMiles = distanceKM / GeographicalPoint.NauticalMilesToKilometres;

					textBoxDistanceKM.Text = GetFormatted(distanceKM, decimalPlaces);
					textBoxDistanceStatuteMiles.Text = GetFormatted(distanceStatuteMiles, decimalPlaces);
					textBoxDistanceNauticalMiles.Text = GetFormatted(distanceNauticalMiles, decimalPlaces);
				}

				{
					double azimuth12 = GeographicalPoint1.GetAzimuth(GeographicalPoint2);
					double azimuth21 = GeographicalPoint2.GetAzimuth(GeographicalPoint1);

					if (Math.Round(azimuth12, decimalPlaces) == 360)
					{
						azimuth12 = 0;
					}

					if (Math.Round(azimuth21, decimalPlaces) == 360)
					{
						azimuth21 = 0;
					}

					textBoxAzimuth12.Text = GetFormatted(azimuth12, decimalPlaces);
					textBoxAzimuth21.Text = GetFormatted(azimuth21, decimalPlaces);
				}
			}
		}

		/// <summary>
		/// Retrieves a normalized directory name.
		/// </summary>
		/// <param name="oldPath">
		/// A normalized or denormalized directory name.
		/// </param>
		/// <param name="withTrailingSeparator">
		/// If true: Append a directory separator character if not there.
		/// If false: Do not.
		/// </param>
		/// <returns>The normalized directory name.</returns>
		static string GetNormalizedPathName(string oldPath, bool withTrailingSeparator)
		{
			string result = null;
			oldPath = oldPath.Trim();

			if (oldPath != null)
			{
				if (oldPath != "")
				{
					char directorySeparatorChar = System.IO.Path.DirectorySeparatorChar;
					int oldLength = oldPath.Length;
					int lastIndex = oldPath.LastIndexOf(directorySeparatorChar);

					if (lastIndex == oldLength - 1)
					{
						if (withTrailingSeparator)
						{
							result = oldPath;
						}
						else
						{
							result = oldPath.Substring(0, oldLength - 1);
						}
					}
					else
					{
						if (withTrailingSeparator)
						{
							result = oldPath + directorySeparatorChar;
						}
						else
						{
							result = oldPath;
						}
					}
				}
				else
				{
					result = oldPath;
				}
			}

			return result;
		}

		static void SetBusyCursor()
		{
			Cursor.Current = Cursors.WaitCursor;
		}

		static void SetDefaultCursor()
		{
			Cursor.Current = Cursors.Default;
		}
	}
}